%  function KernelDimRedTR()
%
%  Computing effective directions for regression with RKHS.
%
%  Author: Kenji Fukumizu (Institute of Statistical Mathematics)
%  Date: March 7, 2006
%
%  (C) Copyright:  Kenji Fukumizu
%
%-----------------------------------------------
% KernelDimRedTR()
%
% Arguments
%  X:  explanatory variables (input data)
%  Y:  response variables (teaching data)
%  K:  dimension of effective subspaces
%  MAX_LOOP: maximum number of iterations
%
% Return value(s)
%  B:  orthonormal column vectors (M x K)
%
% Description
%  This program minimizes the objective function
%     Tr[ Ky(Kz(B)+eps I)^{-1} ] 
%   by the gradient descend method.  
% 
%-----------------------------------------------

function B=KernelDimRedTR(X,Y,K,MAX_LOOP,SG,EPS,ETA,ANL,AA,VERBOSE)

[N,M]=size(X);  % N: data size, M: dim of X.
L=length(Y(1,:));

if K>=M
    fprintf('\nDimension of the effective subspace (K) should be smaller than the dimension of X (M)\n');
    B = 0;
    return;
end

%eps=EPS/N;
eps=EPS;

% Random initialization of projection matrix 
rand('state',1);
seed=floor(rand(1)*100);
randn('state', seed);
B=randn(M,K);
[U,D,V]=svd(B,0);
B=U;

% Orth. projector to the orthogonal complement to <1>
unit=ones(N,N);
I=eye(N);
Q=I-unit./N;
[W,D]=eig(Q);
V=W(:,1:N-1);   % N-1 northonormal vectors orthogonal to <1>

% Gram matrix of Y
aa=sum(Y.*Y,2);
ab=Y*Y';
D=repmat(aa,1,N);
yy=max(D + D' - 2*ab, zeros(N,N));
Gy=exp(-yy./SG);  
Kyo=Q*Gy*Q;
Kyo=(Kyo+Kyo')./2;      % Ky(i,j)=Q*exp(-||y(i)-y(j)||^2/SG)*Q

% Initial value of the objective function
Z=X*B;
nZ=Z./sqrt(SG);
aa=sum(nZ.*nZ,2);
ab=nZ*nZ';
D=repmat(aa,1,N);
ZZ=max(D + D' - 2*ab, zeros(N,N));
Gz=exp(-ZZ);    
Kz=Q*Gz*Q;
Kz=(Kz+Kz')./2;     % Kz(i,j)=Q*exp(-||B^TX(i)-B^TX(j)||^2)*Q

mz=inv(Kz+eps.*I);
tr=trace(Kyo*mz);
if VERBOSE
    fprintf('[0]  trace=%.10f\n', tr);    % Initial value 
end

% Main loop for minimization
XXg=zeros(N*M,N);
for a=1:M
    Xa=repmat(X(:,a),1,N);
    XXg((a-1)*N+1:a*N,:)=Xa-Xa';
end

eta=ETA;
ax=log(AA)/MAX_LOOP;
pdB=zeros(M,K);
for h=1:MAX_LOOP
    sg=SG+(ANL-1)*SG*(MAX_LOOP-h)/MAX_LOOP;     % annealing sg.
    eta=AA*ETA*exp(-ax*h);      % decreasing eta exponentially 
    Z=X*B;                          
    bb=sum(Z.*Z,2);
    ab=Z*Z';
    Kzw=abs(repmat(bb,1,N) + repmat(bb',N,1) - 2*ab);
    Kzw=exp(-Kzw./sg);
    mK=mean(Kzw,2);
    rK=repmat(mK,1,N);
    Kz=Kzw-rK-rK'+mean(mK,1)*unit;        % Kz: Gram matrix  
    Kzi=inv(Kz+EPS.*eye(N));
    
    Ky=exp(-yy./sg);
    mK=mean(Ky,2);
    rK=repmat(mK,1,N);
    Ky=Ky-rK-rK'+mean(mK,1)*unit;   
    Kyzi=Ky*Kzi;
    
    ZZg=zeros(N*K,N);
    for b=1:K
        Zb=repmat(Z(:,b),1,N);
        ZZg((b-1)*N+1:b*N,:)=Zb-Zb';
    end
    for a=1:M
        XX=XXg((a-1)*N+1:a*N,:);
        for b=1:K
            tt=XX.*ZZg((b-1)*N+1:b*N,:).*Kzw;
            mK=mean(tt,2);
            rK=repmat(mK,1,N);
            dKB=tt-rK-rK'+mean(mK,1)*unit;        
            dB(a,b)=trace(Kzi*Kyzi*dKB);
        end
    end
    dB=dB.*(EPS/sg);
    
    % Line search
    mz=inv(Kz+EPS.*I);
    mintr=trace(Ky*mz);
    midx=0;
    nm=norm(dB);
    for i=1:30
        ee=eta*i/(30*nm);
        tmpB=B-ee*dB;
        Z=X*tmpB;
        Kz=reshape(repmat(Z,N,1),N,N,K);
        tmpzz=Kz-permute(Kz,[2,1,3]);
        Kz=Q*exp(-sum(tmpzz.*tmpzz,3)./sg)*Q;  
        mz=inv(Kz+EPS.*I);
        tr=trace(Ky*mz);
        if tr < mintr
            mintr = tr;
            midx=i;
        end
    end
    ee=eta*midx/(30*nm);
    B=B-ee*dB;
    
    [B,D,V]=svd(B,0);
    
    Z=X*B;                          
    Za=reshape(repmat(Z,N,1),N,N,K);
    tmpzz=Za-permute(Za,[2,1,3]);   
    Kz=Q*exp(-sum(tmpzz.*tmpzz,3)./SG)*Q;
    mz=inv(Kz+EPS.*eye(N));
    tr=trace(Kyo*mz);
    if VERBOSE
        fprintf('[%d]  tr=%.10f   eta=%f:%d\n', h, tr, eta, midx);
    end
end




